# <CustomTools>
#  <Menu>
#   	<Submenu name="ImarisOpen">
# 	  <Item name="Counting foci" icon="Python" tooltip="XTension for counting foci from foci and cell users' definitions and making correlation between them">
# 	    <Command>PythonXT::XTCountFoci(%i)</Command>
# 	  </Item>
#   	</Submenu>
#  </Menu>
# </CustomTools>

# authors @LPC-UMR6533-CNRS @IMARIS @henri.Payno @yann.Perrot @lydia.Maigne @emmanuel.Delage
# version v_0.1
# History : 
# 17/09/14 	Beta version allowing users to make foci surfaces in correlation with cell (or nucleus) surfaces in order to generate new statistics.
#			These news statistics include :
#					- mean number of nuclear foci per cell
#					- mean area of nuclear foci (um2) per cell
#					- mean area of nuclear foci (percentage) per cell = sum(fociAreas) / cellArea
# 					- mean number of nuclear foci
# 			Python libraries prerequires :
# 					- python > 2.7
# 					- scipy
# 					- numpy
#					- six
#					- dateutil
#					- pyparsing
#					- matplotlib
#					- pylab
#  			Path to python libraries dateutil, pyparsingPath and sixPath must be registred on the PATH environmental variable 

#			How to run the script: 
#					- Include the script to your XT/python directory. 
# 					- Launch IMARIS
#					- open your image(s)
#					- create two surfaces set. One for nuclear foci areas and one for cell areas (or nucleus areas).
#					- run the script (ImageProcessing/ImarisOpen/XTCountingFoci)
#					- follow instructions displayed on the console user interface.
#
#
#
#
#

import sys
import re
import os
import collections
import operator

####################################################################################################################
# To exit program when displaying an error message
####################################################################################################################
def wait_exit():
	raw_input("\n press enter to exit")
	sys.exit(0)


# get PATH environmental variable 
try:
    user_paths = os.environ['PATH'].split(os.pathsep)
except KeyError:
    user_paths = []

for lPath in user_paths :
	# set them readeable by python sys
	lPath.replace('\\', '/')
	sys.path.append(lPath)

# import ImarisLib library
try:
	import ImarisLib
except ImportError:
	print "Fail to load ImarisLib library"
	wait_exit()

# import scipy
try:
	import scipy
except ImportError:
	print "Fail to load Scipy library"
	wait_exit()

# import numpy
try:
	import numpy
except ImportError:
	print "Fail to load Numpy library"
	wait_exit()

# import six
try:
	import six
except ImportError:
	print "Fail to load the six library"
	wait_exit()

# import dateutil
try:
	import dateutil
except ImportError:
	print "Fail to load dateutil library"
	wait_exit()

# import matplotlib library
try:
	import matplotlib.patches as patches				# used to get the collection.Polygon
	from matplotlib.collections import PolyCollection	# used to get the polygon ang the contain_point function
except ImportError:
	print "Fail to load matplotlib library"
	wait_exit()

# import pylab library
try:
	import pylab
	import matplotlib.pyplot as plt
except ImportError:
	print "Fail to load pylab library"
	wait_exit()

	
# To display results on an other interface 
drawAssociation = False

FociAndCellSurfaces = collections.namedtuple('FociAndCellSurfaces', 'fociSurf cellSurf')
CellStats = collections.namedtuple('CellStats', 'cellIndex cellArea fociSurfacesIndex fociArea fociIntensity')

areaUnit=''
cellsRecoveredByPercentageOfFoci = [10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0]



####################################################################################################################
# 2D cross definition with OA and OB vectors
####################################################################################################################
def cross(o, a, b):
    return (a[0] - o[0]) * (b[1] - o[1]) - (a[1] - o[1]) * (b[0] - o[0])

####################################################################################################################
# return mesh colors according to IDs
####################################################################################################################
def getMeshColor(iMeshId):
	
	if(iMeshId == None):
		return 'r'		# red ( if not associated to no cell )
	if((iMeshId%9) == 0):
		return '0.25'	# 		
	if((iMeshId%8) == 0):
		return 'chartreuse'	# 
	if((iMeshId%7) == 0):
		return 'burlywood'	# 		
	if((iMeshId%6) == 0):
		return '0.75'	# gray
	if((iMeshId%5) == 0):
		return 'm'		# magenta
	if((iMeshId%4) == 0):
		return 'c'		# cyan
	if((iMeshId%3) == 0):
		return 'b'		# blue
	if((iMeshId%2) == 0):
		return 'g'		# green
	
	return 'y'			# yellow

####################################################################################################################
# Definintion of a bounding box
####################################################################################################################
class BoundingBox(object):
	"""class defining a simple bounding box from bottom left and top right points"""
	# constructor
	def __init__(self, mPoints):
		assert(len(mPoints) > 0), "No points given to create the bounding box"
		min_x = mPoints[0][0]
		min_y = mPoints[0][1]
		max_x = min_x
		max_y = min_y
		self.topRight = mPoints[0]
		for iPt in range(0, len(mPoints)):
			if mPoints[iPt][0] > max_x:
				max_x = mPoints[iPt][0]
			if mPoints[iPt][1] > max_y:
				max_y = mPoints[iPt][1]			
			if mPoints[iPt][0] < min_x:
				min_x = mPoints[iPt][0]
			if mPoints[iPt][1] < min_y:
				min_y = mPoints[iPt][1]	

		self.bottomLeft = [min_x, min_y]
		self.topRight = [max_x, max_y]

	# return true if the given point is on the bounding box
	def isIn(self, mPoint):
		return ( ((mPoint[0] <= self.topRight[0]) & (mPoint[0] >= self.bottomLeft[0] )) & ((mPoint[1] <= self.topRight[1]) & (mPoint[1] >= self.bottomLeft[1] )) )

####################################################################################################################
# Definition of meshes from given vertices.
####################################################################################################################
class Mesh(object):
	"""class defing mesh object to compute the position of cell"""
	# constructor
	def __init__(self, mVertices):
		self.vertices = 	mVertices
		self.centroid = 	self.getCentroid()
		self.hull = 		self.getConvexHull(self.vertices)		
		self.boundingBox = 	self.getBoundingBox()

	# return true if the point is in the mesh convex hull
	def isIn(self, mPoint):
		# initially we where supposed to make the calculation using polygons, but calculations failed.
		# in the future, we might use this one, because bounding boxes can give a bad result.
		# if self.boundingBox.isIn(mPoint)==False:
		# 	return False 
		# return self.hull.contains_point([mPoint[0], mPoint[1]])

		return self.boundingBox.isIn(mPoint)

	# return the centroid of the mesh
	def getCentroid(self):
		xSum = 0.
		ySum = 0.
		for iPt in range(0, len(self.vertices)):
			xSum += self.vertices[iPt][0]
			ySum += self.vertices[iPt][1]

		return [xSum/len(self.vertices), ySum/len(self.vertices)]

	# return the bounding box associated to the mesh
	def getBoundingBox(self):
		return BoundingBox(self.vertices)

	# define and return the convex hull of the mesh
	def getConvexHull(self, mVertices):
		# From wikipedia algorithm proposed
	    # """Computes the convex hull of a set of 2D mVertices.
	 
	    # Input: an iterable sequence of (x, y) pairs representing the mVertices.
	    # Output: a list of vertices of the convex hull in counter-clockwise order,
	    #   starting from the vertex with the lexicographically smallest coordinates.
	    # Implements Andrew's monotone chain algorithm. O(n log n) complexity.
	    # """
	    # Boring case: no mVertices or a single point, possibly repeated multiple times.
	    if len(mVertices) <= 1:
	        return mVertices
	 	 
	    # Build lower hull 
	    lower = []
	    for p in mVertices:
	        while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
	            lower.pop()
	        lower.append(p)
	 
	    # Build upper hull
	    upper = []
	    for p in reversed(mVertices):
	        while len(upper) >= 2 and cross(upper[-2], upper[-1], p) <= 0:
	            upper.pop()
	        upper.append(p)
	 
	    # Concatenation of the lower and upper hulls provides the convex hull.
	    # Last point of each list is omitted because it is repeated at the beginning of the other list. 
	    hull = lower[:-1] + upper[:-1]
		    
	    # create a polygon from this hull : 

	    points = []
	    for iPt in range(0, len(hull)):
	    	points.append([ hull[iPt][0], hull[iPt][1] ])

	    return  patches.Polygon(points, closed=True, fill=True)

	# set color
	def setColor(self, pColor):
		self.hull.set_color(pColor)

	# draw the mesh on plt
	def draw(self):
		plt.gca().add_patch(self.hull)
		return

####################################################################################################################
# get foci and cell surfaces from IMARIS with user interface
####################################################################################################################
def getFociAndCellSurfaces(vFactory, vImaris):
	lCellSurf = None
	lFociSurf = None

	while lFociSurf is None:
		raw_input("Select surfaces corresponding to focis and press enter ")
		lFociSurf = vFactory.ToSurfaces(vImaris.GetSurpassSelection())
		# check if selection is a surface
		if lFociSurf is None:
			print( 'The selection isn\'t a surface')

	while lCellSurf is None:
		raw_input("Select surfaces corresponding to cells or nuclei and press enter ")
		lCellSurf = vFactory.ToSurfaces(vImaris.GetSurpassSelection())
		# check if selection is a surface
		if lCellSurf is None:
			print( 'The selection isn\'t a surface')			
		# check if selection is the same surfaces as foci surfaces
		if lCellSurf.GetName() == lFociSurf.GetName():
			print( 'same surfacse as foci surfaces, please pick other surfaces' )
			lCellSurf = None

	return FociAndCellSurfaces(lFociSurf, lCellSurf)

####################################################################################################################
# return cell including foci surface
####################################################################################################################
def getIncludingCell(pPoint, cellMeshes):
	# todo : try the one with the nearest center of mass first
	for iCellMesh in range(0, len(cellMeshes)):
		# to speed up we only make the test of insertion of the centroid of the mesh 
		if cellMeshes[iCellMesh].isIn( pPoint )==True :
			return iCellMesh

	return None

####################################################################################################################
# return true if the given name matches with the IMARIS Area name syntax
####################################################################################################################
def isArea(value):
	return value == 'Area'

####################################################################################################################
# return true if the given name matches with the IMARIS SumIntensity name syntax
####################################################################################################################
def isSumIntensity(value):
	return value == ('Intensity Sum')

####################################################################################################################
# get statistics for cells based on foci features. ( Intensity Sum and Area )
####################################################################################################################
def getCellStats( fociMeshes, cellMeshes, pFociSurfaces, pCellSurfaces):
	fociStats = pFociSurfaces.GetStatistics()
	# get foci areas
	fociAreas = [s for s in range(0, len(fociStats.mNames)) if isArea(fociStats.mNames[s])]
	if(len(fociAreas)>0):
		global areaUnit
		areaUnit = fociStats.mUnits[fociAreas[0]]	
	else:
		print "no area found for foci, make sure you have used an area filter on your surface creation (area must be strored on surfaces stats)"
		wait_exit()

	# get foci sum intensity
	fociSumIntensity = [s for s in range(0, len(fociStats.mNames)) if isSumIntensity(fociStats.mNames[s])]
	if(len(fociSumIntensity) == 0):
		print " Sum intensity not found. Can't process, make sure sum intensity are stored on the foci surfaces statistics tab" 
		wait_exit()

	# get cells areas
	cellStatsFrmIMARIS = pCellSurfaces.GetStatistics()
	cellAreas = [s for s in range(0, len(cellStatsFrmIMARIS.mNames)) if isArea(cellStatsFrmIMARIS.mNames[s])]

	lCellStats = []				# storage of cells stats
	# create stat line for non associated foci 
	unassociatedFoci = CellStats(-1, 0, [], [], [] )
	# create CellStats for eah cell
	for iCell in range(0, len(cellMeshes)):
		# get the area of the iterated cell
		cellArea = [cellStatsFrmIMARIS.mValues[s] for s in cellAreas if (cellStatsFrmIMARIS.mIds[s]==iCell) ]
		if len(cellArea)>0:
			cellArea = cellArea[0]
		else:
			cellArea = 0
			print "No cell area found !!! Do not forget to add an area filter on surfaces to have the area informations !!! "
		
		lCellStats.append(CellStats(iCell, cellArea, [], [], [] ))

	# data foci information in CellStats
	for iMeshFoci in range(0, len(fociMeshes)): 
		cellIncludingFoci = getIncludingCell(fociMeshes[iMeshFoci].centroid, cellMeshes )
		if cellIncludingFoci is None:
			print('\n no corresponding cell found for the foci surfaces of index')
			print fociMeshes[iMeshFoci].centroid
			unassociatedFoci.fociSurfacesIndex.append(iMeshFoci)
		else:
			# add the foci mesh association with the cell mesh
			lCellStats[cellIncludingFoci].fociSurfacesIndex.append(iMeshFoci)
			# add area if registered 
			indexAreaValue = [s for s in fociAreas if (fociStats.mIds[s]==iMeshFoci) ]
			if(len(indexAreaValue)>0):
				lCellStats[cellIncludingFoci].fociArea.append(fociStats.mValues[indexAreaValue[0]])
			else:
				print "no area found for foci, make sure you have used an area filter on your surface creation (area must be strored on surfaces stats)"
				lCellStats[cellIncludingFoci].fociArea.append(0.)
			# add the sum intensity 
			indexSumIntensityValue = [s for s in fociSumIntensity if (fociStats.mIds[s]==iMeshFoci) ]
			if(len(indexSumIntensityValue)>0):
				lCellStats[cellIncludingFoci].fociIntensity.append(fociStats.mValues[indexSumIntensityValue[0]])
			else:
				print "no area found for foci, make sure you have used an area filter on your surface creation (area must be strored on surfaces stats)"
				lCellStats[cellIncludingFoci].fociIntensity.append(0.)
			
		# update mesh color
		fociMeshes[iMeshFoci].setColor(getMeshColor(cellIncludingFoci))
		
	# insert unassociated foci features	at back of lCellsStats
	lCellStats.append(unassociatedFoci)

	return lCellStats

####################################################################################################################
# return where to store the given percentage according to the cellsRecoveredByPercentageOfFoci 
####################################################################################################################
def getIndexFociRecovering(percentage):
	global cellsRecoveredByPercentageOfFoci
	for index in range(0, len(cellsRecoveredByPercentageOfFoci)):
		if percentage < cellsRecoveredByPercentageOfFoci[index]:
			return index

	return len(cellsRecoveredByPercentageOfFoci)-1

####################################################################################################################
# To update the IMARIS 'statistics' tab of cell surfaces
####################################################################################################################
def updateCellStats(vImaris, itemToAddStats, pCellStats, pNbFoci):
	
	listSumFociArea = []
	listIntensitySumFoci = []
	listRelativeSize = []
	listNbFoci = []

	global cellsRecoveredByPercentageOfFoci
	nbCellCellRecovered = [0]*len(cellsRecoveredByPercentageOfFoci)
	# update stats per each cells
	for iStat in range(0, len(pCellStats)-1):

		nbFoci = len(pCellStats[iStat].fociSurfacesIndex )
		areaSum = sum(pCellStats[iStat].fociArea )
		relativeSize = 0
		if pCellStats[iStat].cellArea > 0:
			relativeSize = (( sum(pCellStats[iStat].fociArea ) / pCellStats[iStat].cellArea ) *100.)
			nbCellCellRecovered[getIndexFociRecovering(relativeSize)] += 1

		# insert data to list global statistics
		listRelativeSize.append(relativeSize)
		listSumFociArea.append(areaSum)
		listIntensitySumFoci.append( sum(pCellStats[iStat].fociIntensity) )
		listNbFoci.append(nbFoci)

	# update IMARIS interfaces
	meanRelativeSize = pylab.mean(listRelativeSize)
	stdRelativeSize = pylab.std(listRelativeSize)
	print ('meanRelativeSize +/-std = ' + str(meanRelativeSize) + "% +/- " + str(stdRelativeSize) + " % " )
	itemToAddStats.AddStatistics(['mean relative size'], [meanRelativeSize], ['%'], [['']], [''], [-1] )
	itemToAddStats.AddStatistics(['mean relative size std'], [stdRelativeSize], ['%'], [['']], [''], [-1] )

	meanFociAreaPerCell = pylab.mean(listSumFociArea)
	stdFociAreaPerCell = pylab.std(listSumFociArea)
	print ('mean foci area +/-std = ' + str(meanFociAreaPerCell) + " " + areaUnit  + " +/- " + str(stdFociAreaPerCell) ) + " " + areaUnit
	itemToAddStats.AddStatistics(['mean foci area'], [meanFociAreaPerCell], [areaUnit], [['']], [''], [-1] )
	itemToAddStats.AddStatistics(['foci area std'], [stdFociAreaPerCell], [areaUnit], [['']], [''], [-1] )

	meanNbFoci = pylab.mean(listNbFoci)
	stdNbFoci = pylab.std(listNbFoci);
	print ('meanNbFoci +/-std = ' + str(meanNbFoci) + " +/- " + str(stdNbFoci) )
	itemToAddStats.AddStatistics(['mean nb foci'], [meanNbFoci], ['foci'], [['']], [''], [-1] )
	itemToAddStats.AddStatistics(['mean nb foci std'], [stdNbFoci], ['foci'], [['']], [''], [-1] )

	meanSumIntensity = pylab.mean(listIntensitySumFoci)
	stdDeviationIntensitySum = pylab.std(listIntensitySumFoci)
	print ('meanSumIntensity +/-std = ' + str(meanSumIntensity) + " +/- " + str(stdDeviationIntensitySum) )
	itemToAddStats.AddStatistics(['mean foci intensity sum'], [meanSumIntensity], [''], [['']], [''], [-1] )
	itemToAddStats.AddStatistics(['mean foci intensity sum std'], [stdDeviationIntensitySum], [''], [['']], [''], [-1] )

	print('unassociatedFoci = ' + str(len(pCellStats[len(pCellStats)-1].fociSurfacesIndex)) + " on " + str(pNbFoci) + " focis" )

	# for i in range(0, len(nbCellCellRecovered)) :
	# 	percentage = (nbCellCellRecovered[i]*100.0/(len(pCellStats) - 1) )
	# 	print(percentage)
	# 	print('percentage of nuclei recovered by percentage of foci < ' + str(cellsRecoveredByPercentageOfFoci[i]) + " = " + str( percentage ) )

	return

####################################################################################################################
# convert IMARIS meshes to the Mesh class
####################################################################################################################
def convertToOwnMesh(surfaces):
	meshes = []
	for indexSurface in range(0, surfaces.GetNumberOfSurfaces()):
		meshes.append(Mesh(surfaces.GetVertices(indexSurface)))
	return meshes

####################################################################################################################
# main function
####################################################################################################################
def XTCountFoci(aImarisId):
	if(aImarisId==101):
		print("launching countFoci in debug mode")
	# TODO : send request from : (aSurpassScene.GetChild(i).GetName) ???
	# aSurpassScene.GetNumberOfChildren
	print 'launching the counting foci XTension \n'

	# Create an ImarisLib object
	vImarisLib = ImarisLib.ImarisLib()

	# Get an imaris object with id aImarisId
	vImaris = vImarisLib.GetApplication(aImarisId)

	# Check if the object is valid
	if vImaris is None:
		print 'Could not connect to Imaris!\n'

		# Sleep 2 seconds to give the user a chance to see the printed message
		time.sleep(2)
		return

	# Get the currently loaded dataset  
	vImage = vImaris.GetDataSet()

	# This xtension requires a loaded dataset
	if vImage is None:
		print 'An image has to be loaded into Imaris first\n'
		time.sleep(2)
		return

	# ask user to get the two surfaces we are interested on. One for cells (or nuclei ) and one for focis
	surfacesFociAndCell = getFociAndCellSurfaces(vImaris.GetFactory(), vImaris)

	# convert to our data structure
	fociMeshes = convertToOwnMesh(surfacesFociAndCell.fociSurf)
	cellMeshes = convertToOwnMesh(surfacesFociAndCell.cellSurf)

	# get IMARIS statistics for cells and focis we are interested in
	stats = getCellStats( fociMeshes, cellMeshes, surfacesFociAndCell.fociSurf, surfacesFociAndCell.cellSurf)
	
	# update imaris interface
	updateCellStats(vImaris, surfacesFociAndCell.cellSurf, stats, len(fociMeshes))

	# if requested, draw the correlation made
	if drawAssociation==True:
		# draw each foci with a different color as the cell
		plt.axes()
		# create a CellStats for eah cell
		for iCell in range(0, len(cellMeshes)):
			cellMeshes[iCell].setColor(getMeshColor(iCell))
			cellMeshes[iCell].hull.set_alpha(0.2)
			cellMeshes[iCell].draw()

		for iFoci in range(0, len(fociMeshes)):
			fociMeshes[iFoci].draw()

		plt.axis('scaled')
		pylab.show()

	raw_input("\n press enter to exit")
